==> cpm-1.1.0.tar.gz: cpm-1.1.0/models/lmo.h <==
#ifndef LMO_H_
#define LMO_H_

#include "cpm_models.h"
#include <stdio.h>
#include "mpib.h"

/*!
 * \defgroup lmo LMO: an advanced heterogeneous communication performance model
 * This module provides building of the LMO model and estimation of the execution time of
 * point-to-point and collective communication operations.
 * \{
 */
#ifdef __cplusplus
extern "C" {
#endif

/*! The LMO model */
typedef struct LMO_model {
	/*! Predictor */
	CPM_predictor predictor;
	/*! number of nodes */
	int n;
	/*! array of n average fixed processing delays */
	double* C;
	/*! array of \f$ C_n^2 \f$ average to/from latencies (\f$ L_{ij} = L_{ji} \f$) */
	double* L;
	/*! array of n average variable processing delays */
	double* t;
	/*! array of \f$ C_n^2 \f$ average to/from transmission rates (\f$ \beta_{ij} = \beta_{ji} \f$) */
	double* b;
	/*! threshold between small and large message sizes for the one2many model */
	int S;
	/*! one2many linear model for small message sizes */
	double one2many_small[2];
	/*! one2many linear model for large message sizes */
	double one2many_large[2];
	/*! threshold between small and medium message sizes for the many2one model */
	int M1;
	/*! threshold between medium and large message sizes for the many2one model */
	int M2;
	/*! many2one linear model for small message sizes */
	double many2one_small[2];
	/*! many2one linear model for large message sizes */
	double many2one_large[2];
} LMO_model;

/*!
 * Allocates the LMO model.
 * \param n number of processors
 * \return LMO model
 */
LMO_model* LMO_alloc(int n);

/*!
 * Frees the LMO model.
 * \param model LMO model
 */
void LMO_free(LMO_model* model);

/*! Reads the LMO model. */
void LMO_read(FILE* stream, LMO_model** model);

/*! Writes the LMO model. */
void LMO_write(FILE* stream, const LMO_model* model);

/*!
 * Estimates the point-to-point parameters.
 * \details
 * - Finds the fixed processing delays and latencies.\n
 * For each 3 nodes \f$ i < j < k \f$, measures execution times and solves systems of equations:\n
 * \f$ \begin{cases}
 * T_{ij}(0) = 2 (C_i + L_{ij} + C_j) & i \xlongleftrightarrow[0]{0} j \\
 * T_{jk}(0) = 2 (C_j + L_{jk} + C_k) & j \xlongleftrightarrow[0]{0} k \\
 * T_{ik}(0) = 2 (C_i + L_{ik} + C_k) & i \xlongleftrightarrow[0]{0} k \\
 * T_{ijk}(0) = 2 (2 C_i + \displaystyle\max_{x = j, k}(L_{ix} + C_x)) & i \xlongleftrightarrow[0]{0} jk \\
 * T_{jik}(0) = 2 (2 C_j + \displaystyle\max_{x = i, k}(L_{jx} + C_x)) & j \xlongleftrightarrow[0]{0} ik \\
 * T_{kij}(0) = 2 (2 C_k + \displaystyle\max_{x = i, j}(L_{kx} + C_x)) & k \xlongleftrightarrow[0]{0} ij \\
 * \end{cases} \f$\n
 * averages solutions:\n
 * \f$ T_{ijk}(0) = 2 (2 C_i + \displaystyle\max_{x = j, k}(L_{ix} + C_x)) = 2 C_i + \displaystyle\max_{x = j, k}T_{ix}(0) \f$\n
 * \f$ \begin{cases}
 * C_i = (T_{ijk}(0) - \displaystyle\max_{x = j, k}T_{ix}(0)) / 2 \\
 * C_j = (T_{jik}(0) - \displaystyle\max_{x = i, k}T_{jx}(0)) / 2 \\
 * C_j = (T_{kij}(0) - \displaystyle\max_{x = i, j}T_{kx}(0)) / 2 \\
 * L_{ij} = T_{ij}(0) / 2 - C_i - C_j \\
 * L_{jk} = T_{jk}(0) / 2 - C_j - C_k \\
 * L_{ik} = T_{ik}(0) / 2 - C_i - C_k \\
 * \end{cases} \f$\n
 * and checks confidence intervals.
 * - Finds the variable processing delays and transmission rates.\n
 * For each 3 nodes \f$ i < j < k \f$, solves systems of equations:\n
 * \f$ \begin{cases}
 * T_{ij}(M) = 2 (C_i + L_{ij} + C_j + M (t_i + \displaystyle\frac{1}{\beta_{ij}} + t_j)) & i \xlongleftrightarrow[M]{M} j \\
 * T_{jk}(M) = 2 (C_j + L_{jk} + C_k + M (t_j + \displaystyle\frac{1}{\beta_{jk}} + t_k)) & j \xlongleftrightarrow[M]{M} k \\
 * T_{ik}(M) = 2 (C_i + L_{ik} + C_k + M (t_i + \displaystyle\frac{1}{\beta_{ik}} + t_k)) & i \xlongleftrightarrow[M]{M} k \\
 * T_{ijk}(M) = 2 (2 C_i + M t_i) + \displaystyle\max_{x = j, k}(2 (L_{ix} + C_x) + M (\frac{1}{\beta_{ix}} + t_x)) & i \xlongleftrightarrow[0]{M} jk \\
 * T_{jik}(M) = 2 (2 C_j + M t_j) + \displaystyle\max_{x = i, k}(2 (L_{jx} + C_x) + M (\frac{1}{\beta_{jx}} + t_x)) & j \xlongleftrightarrow[0]{M} ik \\
 * T_{kij}(M) = 2 (2 C_k + M t_k) + \displaystyle\max_{x = i, j}(2 (L_{kx} + C_x) + M (\frac{1}{\beta_{kx}} + t_x)) & k \xlongleftrightarrow[0]{M} ij \\
 * \end{cases} \f$\n
 * averages solutions:\n
 * \f$ T_{ijk}(M) = 2 (2 C_i + M t_i) + \displaystyle\max_{x = j, k}(2 (L_{ix} + C_x) + M (\frac{1}{\beta_{ix}} + t_x)) = 2 C_i + M t_i + \displaystyle\max_{x = j, k}(T_{ix}(0) + T_{ix}(M)) / 2 \f$\n
 * \f$ \begin{cases}
 * t_i = (T_{ijk}(M) - \displaystyle\max_{x = j, k}(T_{ix}(0) + T_{ix}(M)) / 2 - 2 C_i) / M \\
 * t_j = (T_{jik}(M) - \displaystyle\max_{x = i, k}(T_{jx}(0) + T_{jx}(M)) / 2 - 2 C_j) / M \\
 * t_j = (T_{kij}(M) - \displaystyle\max_{x = i, j}(T_{kx}(0) + T_{kx}(M)) / 2 - 2 C_k) / M \\
 * \displaystyle\frac{1}{\beta_{ij}} = (T_{ij}(M) / 2 - C_i - L_{ij} - C_j) / M - t_i - t_j \\
 * \displaystyle\frac{1}{\beta_{jk}} = (T_{jk}(M) / 2 - C_j - L_{jk} - C_k) / M - t_j - t_k \\
 * \displaystyle\frac{1}{\beta_{ik}} = (T_{ik}(M) / 2 - C_i - L_{ik} - C_k) / M - t_i - t_k \\
 * \end{cases} \f$\n
 * and checks confidence intervals.
 * 
 * It is called by \ref LMO_estimate.
 * \param model LMO model, must be allocated and filled by \ref LMO_predict_scatter_flat and \ref LMO_predict_gather_flat (significant only at root)
 * \param comm communicator, number of nodes \f$ \ge 3 \f$
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param precision measurement parameters
 */
void LMO_estimate_p2p(LMO_model* model, MPI_Comm comm, int parallel, MPIB_precision precision);

void LMO_estimate_p2p_(LMO_model* model, MPI_Comm comm, int parallel, MPIB_precision precision);

/*!
 * Estimates the parameters of one-to-many.
 * Measures one-to-many execution time for the message sizes \f$ 0 < M < max\_size \f$ and
 * performs the Bai & Perron algorithm over the F statistic for the data to find the \f$ S \f$ breakpoint
 * in the piecewise linear regression \f$ T \f$ ~ \f$ M \f$. R must be initialized by LMO_init_R at root beforehand.
 * 
 * It is called by \ref LMO_estimate.
 * \param model LMO model, must be allocated (significant only at root)
 * \param comm communicator
 * \param stride stride for message sizes
 * \param max_size maximum message size
 * \param precision measurement parameters
 */
void LMO_estimate_one2many(LMO_model* model, MPI_Comm comm, int stride, int max_size, MPIB_precision precision);

/*!
 * Estimates the parameters of many-to-one.
 * Measures many-to-one execution time for the message sizes \f$ 0 < M < max\_size \f$ and
 * performs the Bai & Perron algorithm over the F statistic for the data to find the \f$ M_2 \f$ breakpoint
 * in the piecewise linear regression \f$ T \f$ ~ 1. R must be initialized by LMO_init_R at root beforehand.
 *
 * Then in the loop measures many-to-one execution time for the message sizes \f$ 0 < M < m \f$
 * with the stride reduced twice each time. \f$ m \f$ is a message size for which a tenfold escalation of the execution time
 * has been observed on the previous step. As stride reaches 1-byte value \f$ m \f$ is truncated to a kb value.
 * 
 * It is called by \ref LMO_estimate.
 * \param model LMO model, must be allocated (significant only at root)
 * \param comm communicator
 * \param stride stride for message sizes
 * \param max_size maximum message size
 * \param precision measurement parameters
 */
void LMO_estimate_many2one(LMO_model* model, MPI_Comm comm, int stride, int max_size, MPIB_precision precision);

/*!
 * Estimates the parameters of the LMO model.
 * Calls \ref LMO_predict_scatter_flat, \ref LMO_predict_gather_flat, \ref LMO_estimate_p2p.
 * R must be initialized by LMO_init_R at root beforehand.
 * \param comm communicator
 * \param precision measurement precision
 * \param msgset message set
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param model LMO model (significant only at root)
 */
void LMO_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
	LMO_model** model);

/*!
 * Predicts the execution time of point-to-point communication.
 * The execution time of \f$ i \xlongrightarrow{M} j \f$ is equal to\n
 * \f$ C_i + L_{ij} + C_j + M (t_i + \displaystyle\frac{1}{\beta_{ij}} + t_j)\f$ 
 * \param _this LMO model
 * \param i index of the precessor
 * \param j index of the precessor
 * \param M message size
 * \return predicted execution time
 */
double LMO_predict_p2p(void* _this, int i, int j, int M);

/*!
 * Predicts the execution time of flat-tree scatter.
 * \f$ (n - 1) (C_r + M t_r) + \displaystyle\max_{i = 1, i \ne r}^{n - 1}
 * (L_{ri} + C_i + M (\frac{1}{\beta_{ri}} + t_i)) \f$ 
 * \param _this LMO model
 * \param root root precessor
 * \param M message size
 * \return predicted execution time
 */
double LMO_predict_scatter_flat(void* _this, int root, int M);

/*!
 * Predicts the execution time of flat-tree gather.
 * \f$ (n - 1) (C_r + M t_r) + \begin{cases}
 * \displaystyle\max_{i = 1}^{n - 1} (L_{ri} + C_i + M (\frac{1}{\beta_{ri}} + t_i)) & M < M_1 \\
 * \displaystyle\sum_{i = 1}^{n - 1} (L_{ri} + C_i + M (\frac{1}{\beta_{ri}} + t_i)) & M > M_2 \\
 * \end{cases} \f$ 
 * \param _this LMO model
 * \param root root precessor
 * \param M message size
 * \return predicted execution time
 */
double LMO_predict_gather_flat(void* _this, int root, int M);

#ifdef __cplusplus
}
#endif
/*!
 * \}
 */

#endif /*LMO_H_*/

==> cpm-1.2.0.tar.gz: cpm-1.2.0/models/lmo.h <==
#ifndef LMO_H_
#define LMO_H_

#include "cpm_model.h"
#include <stdio.h>
#ifdef HAVE_CONFIG_H
#include "MPIBlib/benchmarks/mpib_measurement.h"
#else
#include "mpib_measurement.h"
#endif

/*!
 * \defgroup lmo LMO: an advanced heterogeneous communication performance model
 * This module provides building of the LMO model and estimation of the execution time of
 * point-to-point and collective communication operations.
 * \{
 */
#ifdef __cplusplus
extern "C" {
#endif

/*! \f$ C_n^3 \f$ */
#define LMO_C3(n) (n) * ((n) - 1) * ((n) - 2) / 6

/*!
 * The index of the \f$ (i, j, k) \f$ element in the array of \f$ C_n^3 \f$ elements, \f$ i < j < k < n \f$.
 * \f$ \displaystyle\frac{C_{n - 1}^2 + C_{n - i}^2}{2} i + \frac{(n - i + 1) + (n - j)}{2} (j - i) + (k - j - 1) \f$.
 */
#define LMO_IJK2INDEX(n, i, j, k) (2 * (n) - (i) - 1) * ((i) - 1) * (i) / 4 + (2 * (n) - (i) - (j) + 1) * ((j) - (i)) / 2 + (k) - (j) - 1

/*! The LMO model */
typedef struct LMO_model {
	/*! Predictor */
	CPM_predictor predictor;
	/*! number of nodes */
	int n;
	/*! array of n average fixed processing delays */
	double* C;
	/*! array of \f$ C_n^2 \f$ average to/from latencies (\f$ L_{ij} = L_{ji} \f$) */
	double* L;
	/*! array of n average variable processing delays */
	double* t;
	/*! array of \f$ C_n^2 \f$ average to/from transmission rates (\f$ \beta_{ij} = \beta_{ji} \f$) */
	double* b;
	/*! threshold between small and large message sizes for the one2many model */
	int S;
	/*! one2many linear model for small message sizes */
	double one2many_small[2];
	/*! one2many linear model for large message sizes */
	double one2many_large[2];
	/*! threshold between small and medium message sizes for the many2one model */
	int M1;
	/*! threshold between medium and large message sizes for the many2one model */
	int M2;
	/*! many2one linear model for small message sizes */
	double many2one_small[2];
	/*! many2one linear model for large message sizes */
	double many2one_large[2];
} LMO_model;

/*!
 * Allocates the LMO model.
 * \param n number of processors
 * \return LMO model
 */
LMO_model* LMO_alloc(int n);

/*!
 * Frees the LMO model.
 * \param model LMO model
 */
void LMO_free(LMO_model* model);

/*! Reads the LMO model. */
void LMO_read(FILE* stream, LMO_model** model);

/*! Writes the LMO model. */
void LMO_write(FILE* stream, const LMO_model* model);

/*!
 * Estimates the parameters of the LMO model.
 * -# Measures one-to-many execution time for the message sizes \f$ 0 < M < max\_size \f$ and
 * performs the Bai & Perron algorithm over the F statistic for the data to find the \f$ S \f$ breakpoint
 * in the piecewise linear regression \f$ T \f$ ~ \f$ M \f$.
 * -# Measures many-to-one execution time for the message sizes \f$ 0 < M < max\_size \f$ and
 * performs the Bai & Perron algorithm over the F statistic for the data to find the \f$ M_2 \f$ breakpoint
 * in the piecewise linear regression \f$ T \f$ ~ \f$ 1 \f$.
 * -# In the loop, measures many-to-one execution time for the message sizes \f$ 0 < M < m \f$
 * with the stride reduced twice each time. \f$ m \f$ is a message size for which a tenfold escalation of the execution time
 * has been observed at the previous step. As the stride reaches 1 byte, \f$ m \f$ is truncated to a kb value,
 * which will be \f$ M_1 \f$.
 * -# Finds the fixed processing delays and latencies.\n
 * For each 3 nodes \f$ i < j < k \f$, measures execution times and solves systems of equations:\n
 * \f$ \begin{cases}
 * T_{ij}(0) = 2 (C_i + L_{ij} + C_j) & i \xlongleftrightarrow[0]{0} j \\
 * T_{jk}(0) = 2 (C_j + L_{jk} + C_k) & j \xlongleftrightarrow[0]{0} k \\
 * T_{ik}(0) = 2 (C_i + L_{ik} + C_k) & i \xlongleftrightarrow[0]{0} k \\
 * T_{ijk}(0) = 2 (2 C_i + \displaystyle\max_{x = j, k}(L_{ix} + C_x)) & i \xlongleftrightarrow[0]{0} jk \\
 * T_{jik}(0) = 2 (2 C_j + \displaystyle\max_{x = i, k}(L_{jx} + C_x)) & j \xlongleftrightarrow[0]{0} ik \\
 * T_{kij}(0) = 2 (2 C_k + \displaystyle\max_{x = i, j}(L_{kx} + C_x)) & k \xlongleftrightarrow[0]{0} ij \\
 * \end{cases} \f$\n
 * averages solutions:\n
 * \f$ T_{ijk}(0) = 2 (2 C_i + \displaystyle\max_{x = j, k}(L_{ix} + C_x)) = 2 C_i + \displaystyle\max_{x = j, k}T_{ix}(0) \f$\n
 * \f$ \begin{cases}
 * C_i = (T_{ijk}(0) - \displaystyle\max_{x = j, k}T_{ix}(0)) / 2 \\
 * C_j = (T_{jik}(0) - \displaystyle\max_{x = i, k}T_{jx}(0)) / 2 \\
 * C_j = (T_{kij}(0) - \displaystyle\max_{x = i, j}T_{kx}(0)) / 2 \\
 * L_{ij} = T_{ij}(0) / 2 - C_i - C_j \\
 * L_{jk} = T_{jk}(0) / 2 - C_j - C_k \\
 * L_{ik} = T_{ik}(0) / 2 - C_i - C_k \\
 * \end{cases} \f$\n
 * and checks confidence intervals.
 * -# Finds the variable processing delays and transmission rates.\n
 * For each 3 nodes \f$ i < j < k \f$, solves systems of equations:\n
 * \f$ \begin{cases}
 * T_{ij}(M) = 2 (C_i + L_{ij} + C_j + M (t_i + \displaystyle\frac{1}{\beta_{ij}} + t_j)) & i \xlongleftrightarrow[M]{M} j \\
 * T_{jk}(M) = 2 (C_j + L_{jk} + C_k + M (t_j + \displaystyle\frac{1}{\beta_{jk}} + t_k)) & j \xlongleftrightarrow[M]{M} k \\
 * T_{ik}(M) = 2 (C_i + L_{ik} + C_k + M (t_i + \displaystyle\frac{1}{\beta_{ik}} + t_k)) & i \xlongleftrightarrow[M]{M} k \\
 * T_{ijk}(M) = 2 (2 C_i + M t_i) + \displaystyle\max_{x = j, k}(2 (L_{ix} + C_x) + M (\frac{1}{\beta_{ix}} + t_x)) & i \xlongleftrightarrow[0]{M} jk \\
 * T_{jik}(M) = 2 (2 C_j + M t_j) + \displaystyle\max_{x = i, k}(2 (L_{jx} + C_x) + M (\frac{1}{\beta_{jx}} + t_x)) & j \xlongleftrightarrow[0]{M} ik \\
 * T_{kij}(M) = 2 (2 C_k + M t_k) + \displaystyle\max_{x = i, j}(2 (L_{kx} + C_x) + M (\frac{1}{\beta_{kx}} + t_x)) & k \xlongleftrightarrow[0]{M} ij \\
 * \end{cases} \f$\n
 * averages solutions:\n
 * \f$ T_{ijk}(M) = 2 (2 C_i + M t_i) + \displaystyle\max_{x = j, k}(2 (L_{ix} + C_x) + M (\frac{1}{\beta_{ix}} + t_x)) = 2 C_i + M t_i + \displaystyle\max_{x = j, k}(T_{ix}(0) + T_{ix}(M)) / 2 \f$\n
 * \f$ \begin{cases}
 * t_i = (T_{ijk}(M) - \displaystyle\max_{x = j, k}(T_{ix}(0) + T_{ix}(M)) / 2 - 2 C_i) / M \\
 * t_j = (T_{jik}(M) - \displaystyle\max_{x = i, k}(T_{jx}(0) + T_{jx}(M)) / 2 - 2 C_j) / M \\
 * t_j = (T_{kij}(M) - \displaystyle\max_{x = i, j}(T_{kx}(0) + T_{kx}(M)) / 2 - 2 C_k) / M \\
 * \displaystyle\frac{1}{\beta_{ij}} = (T_{ij}(M) / 2 - C_i - L_{ij} - C_j) / M - t_i - t_j \\
 * \displaystyle\frac{1}{\beta_{jk}} = (T_{jk}(M) / 2 - C_j - L_{jk} - C_k) / M - t_j - t_k \\
 * \displaystyle\frac{1}{\beta_{ik}} = (T_{ik}(M) / 2 - C_i - L_{ik} - C_k) / M - t_i - t_k \\
 * \end{cases} \f$\n
 * and checks confidence intervals.
 * \note R must be initialized by LMO_init_R at root beforehand.
 * \param comm communicator
 * \param precision measurement precision
 * \param msgset message set
 * \param parallel several non-overlapped point-to-point communications at the same time if non-zero
 * \param model LMO model (significant only at root)
 */
void LMO_estimate(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, int parallel,
	LMO_model** model);

/*! Estimates the parameters of the homogeneous LMO model (n = 1) */
void LMO_estimate_homogeneous(MPI_Comm comm, MPIB_precision precision, MPIB_msgset msgset, LMO_model** model);

/*!
 * Predicts the execution time of point-to-point communication.
 * The execution time of \f$ i \xlongrightarrow{M} j \f$ is equal to\n
 * \f$ C_i + L_{ij} + C_j + M (t_i + \displaystyle\frac{1}{\beta_{ij}} + t_j)\f$ 
 * \param _this LMO model
 * \param i index of the processor
 * \param j index of the processor
 * \param M message size
 * \return predicted execution time
 */
double LMO_predict_p2p(void* _this, int i, int j, int M);

/*!
 * Predicts the execution time of flat-tree scatter.
 * \f$ (n - 1) (C_r + M t_r) + \displaystyle\max_{i = 1, i \ne r}^{n - 1}
 * (L_{ri} + C_i + M (\frac{1}{\beta_{ri}} + t_i)) \f$ 
 * \param _this LMO model
 * \param root root processor
 * \param M message size
 * \return predicted execution time
 */
double LMO_predict_scatter_flat(void* _this, int root, int M);

/*!
 * Predicts the execution time of flat-tree gather.
 * \f$ (n - 1) (C_r + M t_r) + \begin{cases}
 * \displaystyle\max_{i = 1}^{n - 1} (L_{ri} + C_i + M (\frac{1}{\beta_{ri}} + t_i)) & M < M_1 \\
 * \displaystyle\sum_{i = 1}^{n - 1} (L_{ri} + C_i + M (\frac{1}{\beta_{ri}} + t_i)) & M > M_2 \\
 * \end{cases} \f$ 
 * \param _this LMO model
 * \param root root processor
 * \param M message size
 * \return predicted execution time
 */
double LMO_predict_gather_flat(void* _this, int root, int M);

#ifdef __cplusplus
}
#endif
/*!
 * \}
 */

#endif /*LMO_H_*/
